#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <math.h>

#include <kmacros.h>
#include <inputline.h>

#include "vcf-format.h"

#define VCF_CHM_ALLOC_STEP (16)
#define VCF_SAMPLE_ALLOC_STEP (32)
#define VCF_SNP_ALLOC_STEP (256*1024)

static char *single_base_strings[4] = { "A", "C", "G", "T" };

static vcf_t *init_vcf(char *vcf_fname) {
  vcf_t *vcf;

  MA(vcf, sizeof(vcf_t));
  vcf->vcf_fname = strdup(vcf_fname);

  FOPEN(vcf->f, vcf_fname, "r", FOPEN_EXIT_ON_ERROR);

  vcf->n_samples = 0;
  vcf->samples = NULL;
  vcf->column_map = NULL;

  vcf->n_subpops = 0;
  vcf->subpop_names = NULL;

  vcf->n_chm = 0;
  vcf->chm = NULL;

  return vcf;
}

static char *scan_field(char *s, char *q) {
  int ql, pl;
  char *p, *qtmp, *rval;

  ql = strlen(q);
  qtmp = alloca(ql + 2);
  strcpy(qtmp, q);
  qtmp[ql] = '=';
  qtmp[ql+1] = 0;
  
  p = strstr(s, qtmp);
  if (p == NULL) return NULL;

  p += ql + 1;
  pl = 0;
  if (p[0] == '<' || p[0] == '"' || p[0] == '\'') {
    pl = 1;
    while(p[pl] != 0 && p[pl] != p[0]) pl++;
    pl -= 1;
    p += 1;
  } else {    
    while(p[pl] != 0 && p[pl] != ',') pl++;
  }

  MA(rval, sizeof(char)*(pl + 1));
  strncpy(rval, p, pl);
  rval[pl] = 0;

  return rval;
}

static void vcf_add_sample(vcf_t *vcf, char *id, char *subpop) {
  int i;

  for(i=0;i<vcf->n_subpops;i++) 
    if (strcmp(subpop, vcf->subpop_names[i])) break;
  if (i == vcf->n_subpops) {
    if (vcf->n_subpops % VCF_SAMPLE_ALLOC_STEP == 0)
      RA(vcf->subpop_names, 
	 sizeof(char *)*(vcf->n_subpops + VCF_SAMPLE_ALLOC_STEP));
    vcf->subpop_names[vcf->n_subpops] = strdup(subpop);
    vcf->n_subpops++;
  }
  
  if (vcf->n_samples % VCF_SAMPLE_ALLOC_STEP == 0)
    RA(vcf->samples, 
       sizeof(sample_t)*(vcf->n_samples + VCF_SAMPLE_ALLOC_STEP));
  
  vcf->samples[vcf->n_samples].sample_name = id;
  vcf->samples[vcf->n_samples].subpop_index = i;
  vcf->samples[vcf->n_samples].column_index = -1;
  vcf->n_samples++;
}

static void vcf_scan_header(vcf_t *vcf, inputline_t *input,
			    char **include_samples, int n_include,
			    char **exclude_samples, int n_exclude) {
  char *inputline, *sample_field, *id, *subpop;
  int i;

  while((inputline = inputline_nextline(input, INPUTLINE_NOCOPY)) != NULL) {
    CHOMP(inputline);
    if (inputline[0] == 0) continue;
    if (strncmp(inputline, "#CHROM", 6) == 0) {
      inputline_pushback(input);
      break;
    }
    if (inputline[0] != '#') break;

    sample_field = scan_field(inputline, "SAMPLE");
    if (sample_field) {
      id = scan_field(sample_field, "ID");
      subpop = scan_field(sample_field, "SUBPOP");

      if (id == NULL) {
	fprintf(stderr,"Warning: SAMPLE field on line %d has no ID tag\n",
		input->line_no);
	if (asprintf(&id, "sample-%d", vcf->n_samples)) ;
      }
      if (subpop == NULL) subpop = strdup("not specified");

      for(i=0;i<n_exclude;i++) if (strcmp(id, exclude_samples[i]) == 0) break;
      if (i < n_exclude) {
	free(id);
	free(subpop);
	free(sample_field);
	continue;
      }

      if (n_include > 0) {
	for(i=0;i<n_include;i++) if (strcmp(id, include_samples[i]) == 0) break;
	if (i == n_include) {
	  free(id);
	  free(subpop);
	  free(sample_field);
	  continue;
	}
      }

      for(i=0;i<vcf->n_samples;i++)
	if (strcmp(vcf->samples[i].sample_name, id) == 0) break;

      if (i < vcf->n_samples) {
	fprintf(stderr,"Warning: sample %s occurs in file more than once.\n",
		id);
	free(id);
      } else {
	vcf_add_sample(vcf, id, subpop);
      }

      free(subpop);
    }
  }

}

static void vcf_map_sample_columns(vcf_t *vcf, char *sample_line) {
  splitfield_t *sf;
  int column_index, sample_index;

  CHOMP(sample_line);
  sf = split(NULL, sample_line, "\t", NULL, SPLIT_NOLIMIT, SPLIT_INLINE);

  MA(vcf->column_map, sizeof(int)*sf->n_fields);
  for(column_index = 0; column_index < 9; column_index++) 
    vcf->column_map[column_index] = -1;

  for(; column_index < sf->n_fields; column_index++) {
    for(sample_index=0; sample_index < vcf->n_samples; sample_index++) 
      if (strcmp(vcf->samples[sample_index].sample_name, 
		 sf->fields[column_index]) == 0) break;

    if (sample_index<vcf->n_samples) {
      vcf->samples[sample_index].column_index = column_index;
      vcf->column_map[column_index] = sample_index;
    } else {
      vcf->column_map[column_index] = -1;
    }
  }

  sf_destroy(sf);
}

static void vcf_map_samples(vcf_t *vcf,
			    char **include_samples, int n_include,
			    char **exclude_samples, int n_exclude,
			    inputline_t *input) {
  char *current_line;

  /* this should return with inputline containing the #CHROM header which 
     defines which columns contain which fields */
  vcf_scan_header(vcf, input, include_samples, n_include, exclude_samples, 
		  n_exclude);
  fprintf(stderr,"VCF header fields defines %d samples in %d groups/subpops\n",
	  vcf->n_samples, vcf->n_subpops);

  /* bork if not */
  current_line = inputline_nextline(input, INPUTLINE_NOCOPY);
  if (strncmp(current_line, "#CHROM", 6) != 0) {
    fprintf(stderr,"Error: #CHROM header line expected in VCF file at line "
	    "%d\n", input->line_no);
    exit(-1);
  }

  vcf_map_sample_columns(vcf, current_line);
}

static int __attribute__((unused))base_toint(char c) {

  switch(c) {
  case 'A':
  case 'a':
    return 0;
  case 'C':
  case 'c':
    return 1;
  case 'G':
  case 'g':
    return 2;
  case 'T':
  case 't':
    return 3;
  case '-':
  default:
    return 4;
  }
}   

static void vcf_input_nextchm(vcf_t *vcf, chm_t *chm, 
			      inputline_t *input, int load_type,
			      double minimum_maf, int minimum_obs,
			      int snp_qthreshold, int gt_qthreshold) {
  char *inputline, *a_allele, *b_allele, gt_tmp[3], *last_format;
  splitfield_t *sf, *format, *gf;
  int i, gt_index, gq_index, column_index, sample_index;
  int split_limit;
  int n_gterror;
  double maf, maf_threshold;
  struct timeval start_time;

  snp_t tsnp;
  int8_t *talleles;
  int8_t *walleles;

  chm->n_snps = 0;
  chm->snps = NULL;
  walleles = NULL;
  MA(talleles, sizeof(int8_t)*vcf->n_samples);

  sf = sf_init();
  format = sf_init();
  gf = sf_init();
  n_gterror = 0;

  gt_tmp[2] = 0;
  gettimeofday(&start_time, NULL);
  last_format = strdup("--- never with match anything bitch ass ---");
  gt_index = gq_index = split_limit = -1;
  while((inputline = inputline_nextline(input, INPUTLINE_NOCOPY)) != NULL) {
    CHOMP(inputline);
    if (inputline[0] == '#' || inputline[0] == 0) continue;

    if (chm->n_snps % 0x1FFF == 0 && chm->n_snps > 0) {
      fprintf(stderr,"\rVCF loading genotypes - chromosome %d - %d SNPs "
	      "/ sec", vcf->n_chm + 1, 
	      (int) ((chm->n_snps / elapsed_time_ms(&start_time))*1000.));
    }

    split(sf, inputline, "\t", "susssdsss", SPLIT_NOLIMIT, SPLIT_INLINE);

    if (strcmp(sf->fields[0], chm->chm_name) != 0) {
      /* save the line which is not part of this chromosome's data inside the 
	 inputline_t object and return to caller */
      inputline_pushback(input);
      break;
    }

    a_allele = sf->fields[3];
    b_allele = sf->fields[4];
    if (strlen(a_allele) > 1 || strlen(b_allele) > 1) {
      if (load_type == VCF_LOAD_SNPS_ONLY) continue;
    } else {
      if (load_type == VCF_LOAD_INDEL_ONLY) continue;
    }

    /* field 5 (counting from 0) is the snp quality score on phred scale */
    if (sf->tf[5].int32 < snp_qthreshold) continue;

    tsnp.pos = sf->tf[1].uint32;
    tsnp.q = sf->tf[5].int32;
    if (strlen(a_allele) == 1) {
      i = 0; while(i<4 && a_allele[0] != single_base_strings[i][0]) i++;
      tsnp.a_allele = (i == 4) ? strdup(a_allele) : 
	single_base_strings[i];
      tsnp.a_int = i;
    } else {
      tsnp.a_allele = strdup(a_allele);
      tsnp.a_int = -1;
    }
    if (strlen(b_allele) == 1) {
      i = 0; while(i<4 && b_allele[0] != single_base_strings[i][0]) i++;
      tsnp.b_allele = (i == 4) ? strdup(b_allele) : 
	single_base_strings[i];
      tsnp.b_int = i;
    } else {
      tsnp.b_allele = strdup(b_allele);
      tsnp.b_int = -1;
    }
    
    /* fields 6, 7, 8 are filter, info, and format respectively, genotype
       fields follow */

    /* decode from the format field where the genotype field is and where the
       genotype quality field is. This is pretty much a waste of CPU time */
    //        if (strcmp(sf->fields[8], last_format) != 0) {
    //          free(last_format);
    //          last_format = strdup(sf->fields[8]);
    
    split(format, sf->fields[8], ":", NULL, SPLIT_NOLIMIT, SPLIT_INLINE);
    gt_index = gq_index = -1;
    i = 0;
    while((gt_index == -1 || gq_index == -1) && i < format->n_fields) {
      if (strcmp(format->fields[i], "GT") == 0) gt_index = i;
      if (strcmp(format->fields[i], "GQ") == 0) gq_index = i;
      i++;
    }
    split_limit = ((gt_index > gq_index) ? gt_index : gq_index) + 1;
      //      fprintf(stderr,"%d,%d,%d\n", gt_index, gq_index, split_limit);
      //	}

    tsnp.nAA = tsnp.nAB = tsnp.nBB = tsnp.n_missing = tsnp.n_obs = 0;
    for(column_index = 9; column_index < sf->n_fields; column_index++) {
      if (vcf->column_map[column_index] == -1) {
	fprintf(stderr,"VCF: Column %d not mapped\n", column_index);
	continue;
      }

      split(gf, sf->fields[column_index], ":", NULL, split_limit, 
	 SPLIT_INLINE);

      sample_index = vcf->column_map[column_index];

      if (gq_index != -1 && atoi(gf->fields[gq_index]) < gt_qthreshold) {
	talleles[sample_index] = GT_NOCALL;
	tsnp.n_missing++;
	continue;
      }


      gt_tmp[0] = gf->fields[gt_index][0];
      gt_tmp[1] = gf->fields[gt_index][2];
      if (strcmp(gt_tmp, "00") == 0) {
	talleles[sample_index] = GT_AA;
	tsnp.nAA++;
      } else if (strcmp(gt_tmp, "11") == 0) {
	talleles[sample_index] = GT_BB;
	tsnp.nBB++;
      } else if (strcmp(gt_tmp, "01") == 0 || strcmp(gt_tmp, "10") == 0) {
	talleles[sample_index] = GT_AB;
	tsnp.nAB++;
      } else if (strcmp(gt_tmp, "..") == 0) {
	talleles[sample_index] = GT_NOCALL;
	tsnp.n_missing++;
      }
      else {
	if (n_gterror < 50) {
	  fprintf(stderr,"genotype %s for sample %s on line %d not "
		  "understood.\n", gf->fields[gt_index], 
		  vcf->samples[sample_index].sample_name, input->line_no);
	} else if (n_gterror == 50) {
	  fprintf(stderr,"Further genotype warnings suppressed.\n");
	}
	talleles[sample_index] = GT_NOCALL;
	tsnp.n_missing++;
	n_gterror++;
      }

    }
    
    tsnp.n_obs = vcf->n_samples - tsnp.n_missing;
    tsnp.a_freq = (tsnp.nAA*2 + tsnp.nAB) / (double) (tsnp.n_obs * 2);
    tsnp.b_freq = (tsnp.nBB*2 + tsnp.nAB) / (double) (tsnp.n_obs * 2);

    maf = (tsnp.a_freq < tsnp.b_freq ? tsnp.a_freq : tsnp.b_freq) *
      tsnp.n_obs;
    maf_threshold = minimum_maf < 1. ? minimum_maf * tsnp.n_obs + 0.5 :
      minimum_maf;

    if (tsnp.n_obs < minimum_obs || maf < maf_threshold) continue;

    if (chm->n_snps % VCF_SNP_ALLOC_STEP == 0) {
      RA(chm->snps, sizeof(snp_t)*(chm->n_snps + VCF_SNP_ALLOC_STEP));
      RA(walleles, sizeof(int8_t)*((chm->n_snps + VCF_SNP_ALLOC_STEP)*
				       vcf->n_samples));
    }

    chm->snps[chm->n_snps] = tsnp;
    memcpy(walleles + chm->n_snps * vcf->n_samples, talleles,
	   sizeof(int8_t)*vcf->n_samples);
    chm->n_snps++;
  }

  MA(chm->alleles, sizeof(int8_t *)*chm->n_snps);
  for(i=0;i<chm->n_snps;i++) chm->alleles[i] = walleles + i*vcf->n_samples;

  free(last_format);
  free(talleles);
}

vcf_t *load_genotypes_vcf(char *vcf_fname, 
			  char **include_samples, int n_include,
			  char **exclude_samples, int n_exclude,
			  int load_type, double minimum_maf, 
			  double minimum_obs, 
			  int snp_qthreshold, 
			  int gt_qthreshold) {
  vcf_t *vcf;
  char *chm_name, *p;
  inputline_t *input;
  int total_snps;
  chm_t tchm;

  vcf = init_vcf(vcf_fname);
  input = init_inputline(vcf->f);
  vcf_map_samples(vcf, include_samples, n_include, exclude_samples, n_exclude, 
		  input);

  if (minimum_obs < 1.) 
    minimum_obs = floor(vcf->n_samples * minimum_obs + 0.5);

  total_snps = 0;
  while((p = inputline_nextline(input, INPUTLINE_NOCOPY)) != NULL) {
    CHOMP(p);
    if (p[0] == '#' || p[0] == 0) continue;
   
    chm_name = strsep(&p, "\t");


    tchm.chm_name = strdup(chm_name);
    tchm.snps = NULL;
    tchm.n_snps = 0;
    tchm.alleles = NULL;

    inputline_pushback(input);
    vcf_input_nextchm(vcf, &tchm, input, load_type, minimum_maf, minimum_obs, 
		      snp_qthreshold, gt_qthreshold);
    if (tchm.n_snps == 0) {
      free(tchm.chm_name);
      continue;
    }

    if (vcf->n_chm % VCF_CHM_ALLOC_STEP == 0)
      RA(vcf->chm, sizeof(chm_t)*(vcf->n_chm + VCF_CHM_ALLOC_STEP));

    vcf->chm[vcf->n_chm] = tchm;
    total_snps += tchm.n_snps;
    vcf->n_chm++;
  }
  fprintf(stderr,"\n");
  fprintf(stderr,"Read %d snps in %d chromosomes/sequences\n", total_snps,
	  vcf->n_chm);

  fclose(vcf->f);
  vcf->f = NULL;

  return vcf;
}

void vcf_free(vcf_t *vcf) {
  int i;

  for(i=0;i<vcf->n_chm;i++) {
    free(vcf->chm[i].alleles[0]);
    free(vcf->chm[i].alleles);
    free(vcf->chm[i].snps);
    free(vcf->chm[i].chm_name);
  }
  free(vcf->chm);

  for(i=0;i<vcf->n_samples;i++)
    free(vcf->samples[i].sample_name);
  free(vcf->samples);
  free(vcf->column_map);
  
  for(i=0;i<vcf->n_subpops;i++)
    free(vcf->subpop_names[i]);
  free(vcf->subpop_names);

  free(vcf->vcf_fname);
  if (vcf->f) fclose(vcf->f);

  free(vcf);
}
