#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <math.h>
#include <stdint.h>
#include <pthread.h>
#include <float.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>

#include <kmacros.h>
#include <cmdline-utils.h>

#include "vcf-format.h"
#include "config.h"

/* command line option global variables */
static char *input_fname, *parents_str, *output_directory;
static double minimum_maf, minimum_obs, impute_threshold;
static int window_length, original_always;
static int use_indel, indel_only, snp_qthreshold, genotype_qthreshold;

/* command line option definitions and descriptions */
static option_t options[] = {
  { 'f', "vcf-input", &input_fname, OPT_STR, 1, 1,
    "VCF format input file to be imputed" },
  { 'm', "minimum-maf", &minimum_maf, OPT_DBL, 0, 1,
    "Minimum minor allele frequency to include for imputation. If <1\n"
    "this value is treated as a proportion. If >1, then this is the minimum\n"
    "number of allele observations." },
  { 'n', "minimum-obs", &minimum_obs, OPT_DBL, 0, 1,
    "Minimum number of observations. If <1, proportion of samples. If >1\n"
    "minimum observation count." },
  { 'w', "window-length", &window_length, OPT_INT, 0, 1,
    "number of flanking SNPs on each side to use for imputation" },

  { 'i', "use-indel", &use_indel, OPT_FLAG, 0, 0,
    "include insertion and deletion polymorphisms" },
  { 'I', "indel-only", &indel_only, OPT_FLAG, 0, 0,
    "use only insertion and deletion polymorphisms (no SNPs)" },
  { 'q', "snp-quality", &snp_qthreshold, OPT_INT, 0, 1,
    "minimum SNP quality to include" },
  { 'g', "genotype-quality", &genotype_qthreshold, OPT_INT, 0, 1,
    "minimum genotype quality to include" },

  { 0, "original-always", &original_always, OPT_FLAG, 0, 0,
    "always output the original genotype if known" },
  { 't', "impute-threshold", &impute_threshold, OPT_DBL, 0, 1,
    "minimum probability to accept an imputation" },
  { 0, "output-directory", &output_directory, OPT_STR, 0, 1,
    "directory to output imputation results" },

  {  0, "parents", &parents_str, OPT_STR, 0, 1,
     "Exactly two parent sample names, separated by a comma, escape spaces" },
  { 0, NULL, NULL, 0, 0, 0, NULL}
};

static char *program_title = 
"GBS missing data inputation exploiting linkage or linkage \n"
  "disequilibrium between nearby markers.";

static char *usage_info = 
"\n"
"Input options:\n"
"  --vcf-input=<filename> or -f <filename>\n"
"    Input GBS genotype data in VCF format\n"
"  --minimum-maf=<number>, -m <number>\n"
"    Minimum minor allele frequency accepted. Argument may be specified as a\n"
"    proportion of observations between 0.0 and 0.5, or as an integer > 1 \n"
"    meaning the minimum count of minor alleles observed\n"
"  --minimum-obs=<number>, -n <number>\n"
"    Minimum number of genotype observations for a marker admitted as input\n"
"    As above, the argument may be a proportion between 0.0 and 1.0 meaning\n"
"    the proportion of input samples observed, or an integer for a raw count\n"
"  --window-length=<integer>, -w <integer>\n"
"    Number of SNPs or markers both 5' and 3' of the impute locus to use as\n"
"    reference loci.\n"
"\n"
"The following options apply only to VCF format input:\n"
"  --use-indel, -i\n"
"    Flag indicating that insertion/deletion markers should be included\n"
"  --indel-only, -I\n"
"    Flag indicating that only in/del markers should be accepted from input\n"
"  --snp-quality=<integer>, -q <integer>\n"
"    Minimum SNP call quality accepted for input.\n"
"  --genotype-quality=<integer>, -g <integer>\n"
"    Minimum genotype call quality to accept as an observation in the input\n"
"\n"
"Output options:\n"
"  --original-always\n"
"    Flag indicating that the original genotype in the input data, where\n"
"    present, should always be output instead of an imputed genotype\n"
"    In some cases, inputation may correct genotype errors\n"
"  --impute-threshold=<decimal>, -t <decimal>\n"
"    Minimum probability an imputed genotype must satisfy to output that\n"
"    genotype. The genotype will otherwise remain as missing data\n"
"  --output-directory=<pathname>  (required)\n"
"    Write output files to the specified directory. Output is currently\n"
"    always in hapmap format for compatibility with downstream uses, even if\n"
"    input was in VCF format. Files present in the directory may be erased.\n"
"\n"
"Notes:\n"
"VCF input format option has not been thoroughly tested on a variety of\n"
"popular programs which produce this format. Current implementation has\n"
"been tested with PANATI only.\n"
"\n"
"Because only hapmap format output is supported at this time, including\n"
"indels from VCF format input may produce incorrect hapmap format files\n"
"as the format to our knowledge does not have a means of expressing in/dels\n"
"or multiple base polymorphisms.\n"
"\n"
"The hapmap format does not allow pass through of SNP call qualities or the\n"
"expression of observed genotype call quality scores or estimated imputation\n"
"quality scores.\n"
"\n"
"*** IMPORTANT: Input is presumed to be in order and physical genome location\n"
"known\n"
 "\n";

 
/* derived cmdline option variables */
static int vcf_load_type;
static char *parent1_name, *parent2_name;
static void init_options(void) {
  input_fname = NULL;

  minimum_maf = 0.1;
  minimum_obs = 0.2;
  window_length = 5;

  use_indel = 0;
  indel_only = 0;
  snp_qthreshold = 50;
  genotype_qthreshold = 20;

  vcf_load_type = VCF_LOAD_SNPS_ONLY;
  parents_str = NULL;

  original_always = 0;
  impute_threshold = 0.8;
  output_directory = "";
}

static void verify_options(void) {
  int stop;

  stop = 0;
  if (input_fname == NULL) {
    fprintf(stderr, "\nSpecify VCF format input file with -f option.");
    stop = 1;
  }

  if (minimum_maf < 0.) {
    fprintf(stderr,"\nOption --minimum-maf (-m) must be non-negative");
    stop = 1;
  }

  if (minimum_obs < 0.) {
    fprintf(stderr,"\nOption --minimim-obs (-n) must be non-negative.");
    stop = 1;
  }

  if (window_length < 1) {
    fprintf(stderr,"\nOption --window-length (-w) must be > 0.");
    stop = 1;
  }

  if (use_indel && indel_only == 0) vcf_load_type = VCF_LOAD_SNPS_AND_INDELS;
  if (indel_only) {
    use_indel = 1;
    vcf_load_type = VCF_LOAD_INDEL_ONLY;
  }

  if (parents_str) {
    char *q1, *q2;

    q1 = strsep(&parents_str, ",");
    q2 = strsep(&parents_str, ",");
    if (q1 == NULL || q2 == NULL) {
      fprintf(stderr,"\nError: Can't interpret --parents option.\n");
      fprintf(stderr,"Specify parent sample names as --parents="
	      "parent1,parent2\n");
      fprintf(stderr,"No space after the comma, enclose in quotes if "
	      "sample names have spaces");
      stop = 1;
    }
    if (strsep(&parents_str,",")) {
      fprintf(stderr,"\nWarning: Specify exactly two parents with --parents "
	      "option.");
      fprintf(stderr,"           samples can not have commas");
    }

    parent1_name = strdup(q1);
    parent2_name = strdup(q2);
  } else {
    parent1_name = NULL;
    parent2_name = NULL;
  }

  if (stop) { 

    fprintf(stderr,"\n\n"
"Program stopped because of command line errors. Run command with no command "
"line options for usage information.\n"); 
    exit(-1); 
  }
}

static double __attribute__((unused))
compute_rsq(int8_t *m1, int8_t *m2, int n_indv,
			  int minimum_obs, int min_allele_count) {
  int i, l;
  double x11, x11_last, D, rsq, p, q, n, N[4][4];
  double __attribute__((unused)) Dp, Dmax; /* D' */

  Dp = -2;
  for(i=0;i<4;i++) {
    for(l=0;l<4;l++) N[i][l] = 0.;
  }

  p = 0.; q = 0.;
  n = 0.;
  for(i=0;i<n_indv;i++) {
    if (m1[i] < 3 && m2[i] < 3) {
      N[m1[i]][m2[i]]++;
      n++;
      p += 2. - m1[i];
      q += 2. - m2[i];
    }
  }
  if (n < minimum_obs) return 0.;
  if (p/2 < min_allele_count || n - p/2 < min_allele_count ||
      q/2 < min_allele_count || n - q/2 < min_allele_count) return 0.;

  p /= 2.*n;
  q /= 2.*n;

  x11 = (2.*N[0][0] + N[0][1] + N[1][0])/(2.*(n - N[1][1]));
  do {
    x11_last = x11;
    x11 = (2.*N[0][0] + N[0][1] + N[1][0] + 
	   N[1][1]*x11_last*(1. - p - q + x11_last)/
	   (x11_last*(1. - p - q + x11_last) + (p - x11_last)*(q - x11_last)))/
      (2.*n);
  } while(fabs(x11_last - x11)>1e-7);
  
  D = x11 - p*q;
  if (D >= 0) {
    Dmax = p*(q-1) < (1-p)*q ? p*(q-1) : (1-p)*q;
  } else {
    Dmax = p*q < (1-p)*(1-q) ? p*q : (1-p)*(1-q);
  }
  Dp = D/Dmax;

  rsq =  D*D/(p*(1-p)*q*(1-q));
  if (rsq > 1.) rsq = 1.;
  return rsq;
}

/* NOTE: this function depends on genotypes codes being 0 - 3, 
   0 = AA, 1 = AB, 2 = BB, 3 = missing. <dp> for diplotype probabilities is
   the return value, and presumed to be an array where the next 9 values 
   correspond to the 9 possible diplotypes */
static double ril_prior[2][2] = { { 0.45, 0.05 }, { 0.05, 0.45 } };
//static double ril_prior[2][2] = { { 0.25, 0.25 }, { 0.25, 0.25 } };
static void estimate_haplotypes_p(double **hp, int8_t *m1, int8_t *m2, 
				  double *p1, double *p2, int n_samples) {
  double haplotype_counts[2][2], pi[2][2], post[2][2], sum, ssq, last_ssq;
  int i, j, itr, nAB;
  double n;

  pi[0][0] = ril_prior[0][0];
  pi[0][1] = ril_prior[0][1];
  pi[1][0] = ril_prior[1][0];
  pi[1][1] = ril_prior[1][1];

  ssq = DBL_MAX;
  itr = 0;
  do {
    last_ssq = ssq;
    haplotype_counts[0][0] =  haplotype_counts[0][1] = 1;
    haplotype_counts[1][0] =  haplotype_counts[1][1] = 1;
    n = 4;
    nAB = 0.;
    for(i=0;i<n_samples;i++) {
      if (m1[i] == 3 || m2[i] == 3) continue;
      
      if (m1[i] == 0) {
	if(m2[i] == 0) {
	  haplotype_counts[0][0] += 2*p1[i]*p2[i];
	} else if (m2[i] == 1) {
	  nAB++;
	  haplotype_counts[0][0] += 1*p1[i]*p2[i];
	  haplotype_counts[0][1] += 1*p1[i]*p2[i];
	} else if (m2[i] == 2) {
	  haplotype_counts[0][1] += 2*p1[i]*p2[i];
	} else {
	  fprintf(stderr,"Unknown diploid genotype code \"%d\"\n", m2[i]);
	  abort();
	}
      } else if (m1[i] == 1) {
	nAB++;
	if(m2[i] == 0) {
	  haplotype_counts[0][0] += 1*p1[i]*p2[i];
	  haplotype_counts[1][0] += 1*p1[i]*p2[i];
	} else if (m1[i] == 1) {
	  nAB++;
	  //	  fprintf(stderr,"Dualing AB genotypes\n");
	  haplotype_counts[0][0] += pi[0][0]*p1[i]*p2[i];
	  haplotype_counts[1][1] += pi[1][1]*p1[i]*p2[i];
	  haplotype_counts[0][1] += pi[0][1]*p1[i]*p2[i];
	  haplotype_counts[1][0] += pi[1][0]*p1[i]*p2[i];
	  continue;
	} else if (m2[i] == 2) {
	  haplotype_counts[0][1] += 1*p1[i]*p2[i];
	  haplotype_counts[1][1] += 1*p1[i]*p2[i];
	} else {
	  fprintf(stderr,"Unknown diploid genotype code \"%d\"\n", m2[i]);
	  abort();
	}
      } else if (m1[i] == 2) {
	if(m2[i] == 0) {
	  haplotype_counts[1][0] += 2*p1[i]*p2[i];
	} else if (m2[i] == 1) {
	  nAB++;
	  haplotype_counts[1][0] += 1*p1[i]*p2[i];
	  haplotype_counts[1][1] += 1*p1[i]*p2[i];
	} else if (m2[i] == 2) {
	  haplotype_counts[1][1] += 2*p1[i]*p2[i];
	} else {
	  fprintf(stderr,"Unknown diploid genotype code \"%d\"\n", m2[i]);
	  abort();
	}
      } else {
	fprintf(stderr,"Unknown diploid genotype code \"%d\"\n", m1[i]);
	abort();
      }
      
      n += p1[i]*p2[i];
    }
    sum = 0.;
    for(i=0;i<2;i++) {
      for(j=0;j<2;j++) {
	post[i][j] = haplotype_counts[i][j] /(double) n * pi[i][j];
	sum += post[i][j];
      }
    }
    
    ssq = 0.;
    for(i=0;i<2;i++) {
      for(j=0;j<2;j++) {
	post[i][j] /= sum;
	ssq += (pi[i][j] - post[i][j])*(pi[i][j] - post[i][j]);
      }
    }
#if 0
    if (post[0][1] > 0.25 || post[1][0] > 0.25 || 
	fabs(post[0][0] - post[1][1]) > 0.333 ) 
      fprintf(stderr, "itr %4d: { %5.3f, %5.3f, %5.3f, %5.3f } n = %1.1f nAB = %d ssq %5.3f\n", itr,
	      post[0][0], post[0][1], post[1][0], post[1][1], n/2., nAB, ssq);
#endif
    itr++;
  } while (fabs(ssq - last_ssq) > 0.001);

  hp[0][0] = post[0][0];
  hp[0][1] = post[0][1];
  hp[1][0] = post[1][0];
  hp[1][1] = post[1][1];

}

static void clean_up(vcf_t *vcf) {
  vcf_free(vcf);
}

typedef struct {
  int pos;
  char *a_allele;
  char *b_allele;

  int n_known;
  int n_missing;
  int n_incorrect;
  int n_imputed;
  double sum_post;
} impute_snp_t;

typedef struct {
  int n_snps;
  impute_snp_t *snps;
  int n_samples;
  char **sample_names;

  int8_t **original_gt;
  int8_t **imputed_gt;
  double **impute_post;
} impute_genotypes_t;

static impute_genotypes_t *vcf_extract_genotypes(vcf_t *vcf, int chm_index) {
  chm_t *chm;
  impute_genotypes_t *gt;
  int i, j, n_samples;

  chm = vcf->chm + chm_index;
  n_samples = vcf->n_samples;

  MA(gt, sizeof(impute_genotypes_t));
  MA(gt->original_gt, sizeof(int8_t *)*chm->n_snps);
  MA(gt->imputed_gt, sizeof(int8_t *)*chm->n_snps);
  MA(gt->impute_post, sizeof(double *)*chm->n_snps);

  MA(gt->original_gt[0], sizeof(int8_t)*chm->n_snps*n_samples);
  MA(gt->imputed_gt[0], sizeof(int8_t)*chm->n_snps*n_samples);
  MA(gt->impute_post[0], sizeof(double)*chm->n_snps*n_samples);

  gt->n_snps = chm->n_snps;
  MA(gt->snps, sizeof(impute_snp_t)*chm->n_snps);

  gt->n_samples = n_samples;
  MA(gt->sample_names, sizeof(char *)*n_samples);
  for(j=0;j<n_samples;j++)
    gt->sample_names[j] = vcf->samples[j].sample_name;

  for(i=0;i<chm->n_snps;i++) {
    gt->snps[i].a_allele = chm->snps[i].a_allele;
    gt->snps[i].b_allele = chm->snps[i].b_allele;
    gt->snps[i].pos = chm->snps[i].pos;
    gt->snps[i].n_known = gt->snps[i].n_missing = 0;
    gt->snps[i].n_incorrect = gt->snps[i].n_imputed = 0;

    gt->original_gt[i] = gt->original_gt[0] + i*n_samples;
    gt->imputed_gt[i] = gt->imputed_gt[0] + i*n_samples;
    gt->impute_post[i] = gt->impute_post[0] + i*n_samples;
  
    for(j=0;j<n_samples;j++) {
      if (chm->alleles[i][j] < -1 || chm->alleles[i][j] > 3) {
	fprintf(stderr,"%d\t%d\t%d\t%d\n", chm_index + 1, i, j, 
		chm->alleles[i][j]);
      }
      gt->original_gt[i][j] = chm->alleles[i][j] != -1 ? chm->alleles[i][j] :3;
      gt->imputed_gt[i][j] = gt->original_gt[i][j];
      gt->impute_post[i][j] = gt->original_gt[i][j] != 3 ? 0.95 : 0.0;
    }
  }


  return gt;
}

static void impute_genotypes_free(impute_genotypes_t *gt) {

  free(gt->original_gt[0]);
  free(gt->imputed_gt[0]);
  free(gt->impute_post[0]);

  free(gt->original_gt);
  free(gt->imputed_gt);
  free(gt->impute_post);

  free(gt->sample_names);
  free(gt->snps);

  free(gt);

}

static int __attribute__((unused)) predict_genotype(double *r_post, double **hp, int ref_gt) {
  double p[3], s;
  int max;

  if (ref_gt == 3) {
    *r_post = 0.;
    return 3;
  }

  if (ref_gt == 0) {
    p[0] = hp[0][0] * hp[0][0];
    p[1] = hp[0][0] * hp[0][1];
    p[2] = hp[0][1] * hp[0][1];
  } else if (ref_gt == 1) {
    p[0] = hp[0][0] * hp[1][0];
    p[1] = hp[0][0] * hp[1][1] + hp[0][1] * hp[1][0];
    p[2] = hp[0][1] * hp[1][1];
  } else if (ref_gt == 2) {
    p[0] = hp[1][0] * hp[1][0];
    p[1] = hp[1][0] * hp[1][1];
    p[2] = hp[1][1] * hp[1][1];
  } else {
    fprintf(stderr,"Unknown diploid genotype code \"%d\"\n", ref_gt);
    abort();
  }
  s = p[0] + p[1] + p[2];
  p[0] /= s;
  max = 0;
  p[1] /= s;
  if (p[1] > p[0]) max = 1;
  p[2] /= s;
  if (p[2] > p[max]) max = 2;

  *r_post = p[max];
  return max;
}

static void calculate_genotype_p(double *p, double **hp, int ref_gt) {

  if (ref_gt == 3) return;

  if (ref_gt == 0) {
    p[0] *= hp[0][0] * hp[0][0];
    p[1] *= hp[0][0] * hp[0][1];
    p[2] *= hp[0][1] * hp[0][1];
  } else if (ref_gt == 1) {
    p[0] *= hp[0][0] * hp[1][0];
    p[1] *= hp[0][0] * hp[1][1] + hp[0][1] * hp[1][0];
    p[2] *= hp[0][1] * hp[1][1];
  } else if (ref_gt == 2) {
    p[0] *= hp[1][0] * hp[1][0];
    p[1] *= hp[1][0] * hp[1][1];
    p[2] *= hp[1][1] * hp[1][1];
  } else {
    fprintf(stderr,"Unknown diploid genotype code \"%d\"\n", ref_gt);
    abort();
  }

}
  

static int impute_genotypes(impute_genotypes_t *gt, char *chm_label) {
  int i, j, k;
  double ***hp, p[3], sum, sum_post;
  int max, nAB_switched;
  int n_missing, n_imputed, n_known, n_attempted, n_incorrect, n_changed;

  hp = alloca(window_length*2*sizeof(double **));
  for(i=0;i<window_length*2;i++) {
    hp[i] = alloca(sizeof(double *)*2);
    hp[i][0] = alloca(sizeof(double)*2);
    hp[i][1] = alloca(sizeof(double)*2);
  }

  fprintf(stderr,"Imputing %s ... ",chm_label);
  for(i=0;i<gt->n_snps;i++) {
    for(j=0;j<gt->n_samples;j++) {
      if (gt->original_gt[i][j] != 3) {
	gt->imputed_gt[i][j] = gt->original_gt[i][j];
	gt->impute_post[i][j] = 0.95;
      }
    }
    gt->snps[i].n_known = gt->snps[i].n_incorrect = 0;
    gt->snps[i].n_missing = gt->snps[i].n_imputed = 0;
    gt->snps[i].sum_post = 0.;
  }

  n_changed = 0;
  n_missing = n_imputed = n_known = n_attempted = n_incorrect = 0;
  sum_post = 0.;
  for(i=0;i < gt->n_snps;i++) {
    for(k=0;k<window_length;k++) {
      if (i-k-1 >= 0) {
	estimate_haplotypes_p(hp[k], gt->imputed_gt[i-k-1],
			      gt->imputed_gt[i], gt->impute_post[i-k-1],
			      gt->impute_post[i], gt->n_samples);
      }
      if (i+k+1 < gt->n_snps) {
	estimate_haplotypes_p(hp[k + window_length], gt->imputed_gt[i+k+1],
			      gt->imputed_gt[i], gt->impute_post[i+k+1],
			      gt->impute_post[i], gt->n_samples);
      }
    }
    
    nAB_switched = 0;
    for(j=0;j<gt->n_samples;j++) {
      p[0] = 0.475;
      p[1] = 0.05;
      p[2] = 0.475;
      for(k=0;k<window_length;k++) {
	if (i-k-1 > 0)
	  calculate_genotype_p(p, hp[k], gt->imputed_gt[i-k-1][j]);
	if (i+k+1 < gt->n_snps)
	  calculate_genotype_p(p,hp[k+window_length],gt->imputed_gt[i+k+1][j]);
      }
      sum = p[0] + p[1] + p[2];
      p[0] /= sum;
      max = 0;
      p[1] /= sum;
      if (p[1] > p[0]) max = 1;
      p[2] /= sum;
      if (p[2] > p[max]) max = 2;

      if (gt->original_gt[i][j] == 3) {
	gt->snps[i].n_missing++;
	n_missing++;
	if (p[max] > impute_threshold) {
	  if (gt->imputed_gt[i][j] != max) n_changed++;
	  gt->imputed_gt[i][j] = max;
	  gt->impute_post[i][j] = p[max];
	  gt->snps[i].n_imputed++;
	  n_imputed++;
	}
      } else {
	gt->snps[i].n_known++;
	n_known++;
	if (gt->original_gt[i][j] != max) {
	  gt->snps[i].n_incorrect++;
	  n_incorrect++;
	  if (gt->original_gt[i][j] == 1) nAB_switched++;
	}
      }
      gt->snps[i].sum_post += p[max];
      sum_post += p[max];
      n_attempted++;
    }

    //    fprintf(stderr,"For marker pair (%5d,%5d) - %3d known AB switched. %5.1f%% missing %5.1f%% imputed %5.1f%% incorrect\n", i, i+1, nAB_switched, n_missing/(double) gt->n_samples*100., n_imputed/(double) (n_missing + 1e-15) * 100., n_incorrect/(double) n_known * 100.);
  }
  fprintf(stderr,"%4.1f%% missing, %5.1f%% imputed, "
	  "Accuracy estimate %5.1f%%, posterior sum %1.1f%%\n", 
	  n_missing/(double) n_attempted * 100., 
	  n_imputed/(double) n_missing * 100., 
	  (1. - n_incorrect/(double) n_known) * 100.,
	  sum_post/n_attempted * 100.);

  return n_changed;
}

static char *gt_string(int gt, char *a_allele, char *b_allele) {
  switch(gt) {
  case 0:
    return a_allele;
  case 1:
    if (strlen(a_allele)>1 || strlen(b_allele)>1) return "H";
    switch(a_allele[0]) {
    case 'A':
      switch(b_allele[0]) {
      case 'A': return "A";
      case 'C': return "M";
      case 'G': return "R";
      case 'T': return "W";
      default: return "N";
      }
    case 'C':
      switch(b_allele[0]) {
      case 'A':return "M";
      case 'C': return "C";
      case 'G': return "S";
      case 'T': return "Y";
      default: return "N";
      }
    case 'G':
      switch(b_allele[0]) {
      case 'A': return "R";
      case 'C': return "S";
      case 'G': return "G";
      case 'T': return "K";
      default: return "N";
      }
    case 'T':
      switch(b_allele[0]) {
      case 'A': return "W";
      case 'C': return "Y";
      case 'G': return "K";
      case 'T': return "T";
      default: return "N";
      }
    }
    return "N";
  case 2:
    return b_allele;
  case 3:
    return "N";
  default:
    fprintf(stderr,"Unknown genotype code %d in gt_string()\n", gt);
    return "N";
  }
}

#define TMP_STR_MAX (8192)
static void plumage_output(impute_genotypes_t *gt, int chm_index,
			   char *chm_name) {
  char fname[TMP_STR_MAX], id_str[TMP_STR_MAX];
  impute_snp_t *snps;
  FILE *f;
  DIR *d;
  int i, j, k, n_fail;

  snps = gt->snps;

  if (strlen(output_directory) > 0) {
    if ((d = opendir(output_directory)) == NULL) {
      if (errno == ENOENT) {
	if (mkdir(output_directory, 0777) == -1) {
	  fprintf(stderr,"Output directory %s does not exist and can not be "
		  "created (%s)\n", output_directory, strerror(errno));
	  exit(-1);
	}
      } else {
	fprintf(stderr,"Can't access output directory %s", output_directory);
	exit(-1);
      }
    } else {
      closedir(d);
    }
    snprintf(fname, TMP_STR_MAX, "%s/imputation.final.output.chr%d.tsv",
	     output_directory, chm_index + 1);
  } else {
    snprintf(fname, TMP_STR_MAX, "imputation.final.output.chr%d.tsv",
	     chm_index + 1);
  }

  f = fopen(fname,"w");
  if (f == NULL) {
    fprintf(stderr,"Error: Can't open output file %s (%s) - output skipped\n",
	    fname, strerror(errno));
    return;
  }

  /* write column headers */
  fprintf(f,"\tsnp_id\talleles\tchr\tpos");
  for(j=0;j<gt->n_samples;j++)
    fprintf(f,"\t%s", gt->sample_names[j]);
  fprintf(f,"\timputation.note\taccuracy estimate\tpre-impute-missing\tpost-impute-missing\n");

  /* output genotypes */
  n_fail = 0;
  for(i=0;i<gt->n_snps;i++) {
    if (snps[i].sum_post/gt->n_samples < impute_threshold ||
	1. - (gt->n_samples - snps[i].n_known - snps[i].n_imputed)/(double) gt->n_samples < impute_threshold) {
      n_fail++;
      continue;
    }

    /* check if marker has same position has adjacent ones. If so, add a suffix
       to the snp_id that is output */
    k = i;
    while(k>=0 && snps[i].pos == snps[k].pos) k--;
    if (i-k-1 > 0 || (i<gt->n_snps-1 && snps[i].pos == snps[i+1].pos)){
      snprintf(id_str, TMP_STR_MAX, "%d.%d", snps[i].pos, i-k-1);
    } else {
      snprintf(id_str, TMP_STR_MAX, "%d", snps[i].pos);
    }

    /* leading columns with SNP/marker definition */
    fprintf(f,"%d\tS%d_%s\t%s/%s\t%d\t%d", i+1, chm_index + 1, id_str, 
	    snps[i].a_allele, snps[i].b_allele, chm_index + 1, snps[i].pos);

    /* genotype output proper */
    for(j=0;j<gt->n_samples;j++) {

      /* if an original genotype exists and the user has requested that these 
	 always be used in the output (cmdline), or if the imputed genotype 
	 does not meet the impute_threshold set by the user, output the original
	 genotype */
      if ((original_always || gt->impute_post[i][j] < impute_threshold) 
	  && gt->original_gt[i][j] != 3) {
	fprintf(f,"\t%s", gt_string(gt->original_gt[i][j], snps[i].a_allele,
				    snps[i].b_allele));
      } else {
	/* output imputed genotype if it meets the user specified threshold */
	if (gt->impute_post[i][j] >= impute_threshold) {
	  fprintf(f,"\t%s", gt_string(gt->imputed_gt[i][j], snps[i].a_allele,
				      snps[i].b_allele));
	} else {
	  //	  fprintf(f,"\tNaN_CG");
	  fprintf(f,"\tNaN_%s%s", snps[i].a_allele, snps[i].b_allele);
	}
      }
    }

    /* Information about imputation results */
    fprintf(f,"\t%1.3f\t%1.3f\t%1.3f\t%1.3f\n", 
	    1 - snps[i].n_incorrect/(double) (snps[i].n_known + 1e-15),
	    1 - snps[i].n_incorrect/(double) (snps[i].n_known + 1e-15),
	    snps[i].n_missing/(double) gt->n_samples,
	    (gt->n_samples - snps[i].n_known - snps[i].n_imputed)/(double) gt->n_samples);
  }
  fprintf(stderr,"chm %s - %d SNPs posterior sum below threshold\n", chm_name,
	  n_fail);
    
	  
}


static void do_imputation(vcf_t *vcf) {
  impute_genotypes_t *gt;
  int i;

  for(i=0;i<vcf->n_chm && i<12;i++) {
    gt = vcf_extract_genotypes(vcf, i);
    impute_genotypes(gt, vcf->chm[i].chm_name);
    while(impute_genotypes(gt, vcf->chm[i].chm_name)>0) ;
    plumage_output(gt, i, vcf->chm[i].chm_name);
    impute_genotypes_free(gt);
  }
}

static void standardize_genotypes(vcf_t *vcf, char *parent1_name,
				  char *parent2_name) {
  int i, j, k, l;
  int *remove_snps, n_remove, total_removed, n_remaining, original_nsnps;
  int parent1_index, parent2_index;
  int n_keep, n_swap;
  __attribute__((unused))char *stmp;
  __attribute__((unused))int itmp;
  __attribute__((unused))int8_t atmp;
  __attribute__((unused))double dtmp;

  fprintf(stderr,"Standardizing genotypes... ");
  for(i=0;i<vcf->n_samples;i++) 
    if (strcmp(vcf->samples[i].sample_name, parent1_name) == 0) break;
  if (i == vcf->n_samples) {
    fprintf(stderr,"Error: Can't find parent sample %s in genotype input.\n",
	    parent1_name);
    exit(-2);
  } else {
    fprintf(stderr,"Parent sample %s found at index %d\n", parent1_name, i);
  }
  parent1_index = i;

  for(i=0;i<vcf->n_samples;i++) 
    if (strcmp(vcf->samples[i].sample_name, parent2_name) == 0) break;
  if (i == vcf->n_samples) {
    fprintf(stderr,"Error: Can't find parent sample %s in genotype input.\n",
	    parent2_name);
    exit(-2);
  } else {
    fprintf(stderr,"Parent sample %s found at index %d\n", parent2_name, i);
  }
  parent2_index = i;

  original_nsnps = total_removed = n_remaining = 0;
  for(i=0;i<vcf->n_chm;i++) {
    chm_t *chm;
    chm = vcf->chm + i;

    original_nsnps += chm->n_snps;
    remove_snps = NULL;
    n_remove = 0;
    n_keep = n_swap = 0;
    for(j=0;j<chm->n_snps;j++) {
      if (chm->alleles[j][parent1_index] == 0 &&
	  chm->alleles[j][parent2_index] == 2) {
	/*	fprintf(stderr,"Keeping alleles for SNP %d at chm %3d pos %11d\n", \
		j, i, chm->snps[j].pos); */
	n_keep++;
	continue;
      }

      if (chm->alleles[j][parent1_index] != 2 ||
	  chm->alleles[j][parent2_index] != 0) {
	fprintf(stderr,"Remove SNP %d\n", j);
	if (n_remove % 256 == 0) 
	  RA(remove_snps, sizeof(int)*(n_remove + 256));
	remove_snps[n_remove++] = j;
	continue;
      }

      n_swap++;
      /*fprintf(stderr,"Swapping alleles for SNP %d at chm %3d pos %11d\n", 
	      j, i, chm->snps[j].pos);*/
      stmp = chm->snps[j].a_allele;
      chm->snps[j].a_allele = chm->snps[j].b_allele;
      chm->snps[j].b_allele = stmp;
	
      itmp = chm->snps[j].nAA;
      chm->snps[j].nAA = chm->snps[j].nBB;
      chm->snps[j].nBB = itmp;
	
      atmp = chm->snps[j].a_int;
      chm->snps[j].a_int = chm->snps[j].b_int;
      chm->snps[j].b_int = atmp;
	
      dtmp = chm->snps[j].a_freq;
      chm->snps[j].a_freq = chm->snps[j].b_freq;
      chm->snps[j].b_freq = dtmp;

      for(k=0;k<vcf->n_samples;k++) {
	if (chm->alleles[j][k] == 2) {
	  chm->alleles[j][k] = 0;
	} else if (chm->alleles[j][k] == 0) {
	  chm->alleles[j][k] = 2;
	}
      }

    }

    fprintf(stderr,"%d SNPs kept, %d (%1.1f%%) SNPs swapped on chm %d\n", 
	    n_keep, n_swap, n_swap*100./chm->n_snps, i+1);
    fprintf(stderr,"%d SNPs to remove for chm %d\n",n_remove,i);
    if (n_remove > 0) {
      j = 0;
      k = 0;
      l = 0;
      if (remove_snps[l] == 0) {
	fprintf(stderr,"Want to remove the first SNP!\n");
	l = 1;
      }
      while(l<n_remove) {
	while(j<chm->n_snps && j != remove_snps[l]) {
	  chm->alleles[k] = chm->alleles[j];
	  chm->snps[k] = chm->snps[j];
	  j++;
	  k++;
	}
	j++;
	l++;
      }
      while(j<chm->n_snps) {
	chm->alleles[k] = chm->alleles[j];
	chm->snps[k] = chm->snps[j];
	j++;
	k++;
      }
      chm->n_snps = k;
      total_removed += n_remove;

      if (remove_snps) free(remove_snps);
    }
    n_remaining += chm->n_snps;
  }

  fprintf(stderr,"%d SNPs dropped, %d (%1.1f%%) SNPs remain.\n",
	  total_removed, n_remaining, n_remaining*100./original_nsnps);
}

int main(int argc, char *argv[]) {
  vcf_t *vcf;

  if (argc == 1) {
    fprintf(stderr,"\n%s - %s\n", PACKAGE_STRING, program_title);
    fprintf(stderr,"%s", usage_info);
    exit(0);
  }

  init_options();
  cmdline_getoptions(options, argc, argv);
  verify_options();

  vcf = load_genotypes_vcf(input_fname, NULL, 0, NULL, 0, vcf_load_type,
			   minimum_maf, minimum_obs,
			   snp_qthreshold, genotype_qthreshold);
  if (parent1_name)
    standardize_genotypes(vcf, parent1_name, parent2_name);

  do_imputation(vcf);

  clean_up(vcf);
  return 0;
}
